package prodcons02;
public class Main{
    private Usuario[] usr;     //  Usuarios
    private PrintServ[] prs;    //  Impresoras

    public Main() {
        usr = new Usuario[10];   //  10 usuarios generando documentos
        prs = new PrintServ[2];   //  2 impresoras  imprimiendo
    }   
    public static void main(String[] args){
        Main main = new Main();
        main.demo();
    }
    void demo(){
        Cola cola = new Cola();     //  Generamos una cola de impresin
        for(int j=0;j<2;j++){
            prs[j] = new PrintServ(cola);   //  Definimos  2 impresoras
            prs[j].start();             //  las arrancamos ...
        }
        String aux;
        for(int j=0;j<10;j++){
            aux="0"+j;
            usr[j] = new Usuario(cola, aux); //  Definimos 10 usuarios
            usr[j].start();             //  y los ponemos a trabajar
        }
    }   //  void demo()
}       //  class Main

class Cola{             //  Oficia de monitor
    private Celda primero, ultimo;  //  Punteros al primer y ultimo documento
    private static int tCon;        //  total impreso
    private static int pos;         //  Posicion en el texto
    private static String texto;    //  texto a tratar
    Cola(){    // constructor sin argumentos
        primero=ultimo=null;
        tCon = 0;
        pos = 0;
        texto = "Durante dos das sopl un viento clido del Sur.";
        texto+= "La primavera azul se apoder de la vasta estepa del Don...";
    }/* Los mtodos estticos sincronizados endFila() y contSes() 
    implementan bloqueo a nivel clase. Este bloqueo en necesario para
    conseguir imprimir los cuarenta documentos que generan 10 usuarios  
    y que deben ser impresos a razn de 10 p/sesion (o renglon)     */

    public static synchronized boolean endFila(){  // Fin de fila o rengln 
        if(++tCon%10==0) return true;
        else return false;
    }

    public static synchronized boolean contSes(){  //  Contina la sesin ? 
        if(pos+4 < texto.length()) return true;
        else return false;
    }

    /* Los mtodos de instancia sincronizados inclUlt() (Incluir ltimo) 
     y exclPri() (Excluir primero) implementan bloqueo a nivel instancia.
    El bloqueo es necesario para conseguir gestionar adecuadamente un nico
    objeto Cola que recibe textos de 10 hilos productores (Clase Usuario) 
    y los descarga en 2 hilos consumidores (clase printServ)        */

    public synchronized void inclUlt(String nome) throws NullPointerException{
        String doc = "("+nome+") "; 
        doc+= texto.substring(pos,pos+4)+" ";
        if(primero==null) primero=ultimo=new Celda(doc,null); 
        else ultimo=ultimo.siguiente=new Celda (doc,null);
        pos+=4;
        notifyAll();       
        return;
    }

    public synchronized void exclPri() throws InterruptedException{ 
        String textImp = "**************";  //  texto a imprimir si error
        while (primero==null && contSes())   // No tenemos texto en cola, pero
            wait();         // la sesin no termin, esperemos que entre algo
        if(primero != null){     // Ya entro texto en cola
            textImp = primero.elemento;
            if(primero.equals(ultimo)) // Slo un texto en cola
                primero=ultimo=null;
            else primero=primero.siguiente;
            System.out.print(textImp);
            notifyAll();       
        }
    }   //  exclPri()
}       // class

class Celda{ 
    Celda siguiente;
    String elemento;
    public Celda(String elemento){ 
        this.elemento = elemento;
    }
    public Celda(String elemento, Celda siguiente){ 
        this.elemento= elemento;
        this.siguiente=siguiente;
    }

    public String getElemento(String elemento){return elemento;}
    public void setElemento(String elemento){ this.elemento=elemento;}
    public Celda getSiguiente(){return siguiente;}
    public void setSiguiente(){this.siguiente=siguiente;}
}

// PRODUCTOR 
class Usuario extends Thread{ // Estaciones generan spool 
    private Cola cola;
    private String nome;
    public Usuario(Cola cola, String nome){ // Constructor 
         this.cola = cola;
         this.nome = nome;
    }
    public void run (){
        while(Cola.contSes()){
            try{
                cola.inclUlt(nome); // mtodo de instancia sincronizado
                sleep((int)(1000*Math.random()));
            }catch (NullPointerException e){ 
                System.err.println("InclUlt() :" + e.getMessage());
            }catch (InterruptedException e){ 
                System.err.println("Int. Except. " + e.getMessage());
            }
        }   //  while
    }       //  run()
}   //  class Usuario

// CONSUMIDOR 
class PrintServ extends Thread{ // Servidores de impresin atienden spool  
    private Cola cola;
    PrintServ(Cola cola){this.cola = cola;}

    public void run(){
        while(Cola.contSes()){
            try{
                cola.exclPri();
                if(Cola.endFila())System.out.println(); //  Avanzamos renglon
            }catch (InterruptedException e){ 
                System.err.println("exclPri(): " + e.getMessage());
            }
        }   //  while
    }   //  void run()
}      //  class printServ
